/* @applitools/dom-snapshot@4.4.8 */
function __pollResult() {
  var pollResult=function(){"use strict";return function(){const t=function(t){return function(){function t(t,n=0){const e=t.charCodeAt(n);return e>=55296&&e<56320?1024*(e-55296)+(t.charCodeAt(n+1)-56320)+65536:56320<=e&&e<=57343?-1:e}var n=function(n,e){const o=[];let r=0;for(let u=0;u<n.length;++u){const c=t(n,u);let a=0;c>0&&(a=c<128?1:c<2048?2:c<65536?3:c<2097152?4:c<67108864?5:6),r+a>e?(o.push(u),r=a):r+=a}return o};const e="SUCCESS",o="SUCCESS_CHUNKED",r="ERROR";var u=function(t,u,c={}){const a=function(t,{chunkByteLength:u=0}={}){if(t){if(t.value){if(u){if(!t.chunks){const e=JSON.stringify(t.value);t.chunks=n(e,u),t.chunks.length>0&&(t.from=0,t.value=e)}if(t.from>=0)return{status:o,value:t.value.substring(t.from,t.from=t.chunks.shift()),done:!t.from}}return{status:e,value:t.value}}return t.error?{status:r,error:t.error}:{status:"WIP"}}return{status:r,error:"unexpected poll request received - cannot find state of current operation"}}((t=t||{})[u],c);return(a.status===e||a.status===r||a.status===o&&a.done)&&(t[u]=null),a},c={chunkify:n,pollify:function(t,n,e){return o=>function(){return n[e]||(n[e]={},t.apply(null,arguments).then(t=>n[e].value=t).catch(t=>n[e].error=t.message)),u(n,e,o)}},poll:u,absolutizeUrl:function(t,n){return new URL(t,n).href},isInlineFrame:function(t){return t.contentDocument&&t.contentDocument.location&&!/^https?:$/.test(t.contentDocument.location.protocol)},isAccessibleFrame:function(t){try{const n=t.contentDocument;return Boolean(n&&n.defaultView&&n.defaultView.frameElement)}catch(t){return!1}}};const{poll:a}=c,{EYES_NAMESPACE:s,DOM_SNAPSHOT_KEY:l}={EYES_NAMESPACE:"__EYES__APPLITOOLS__",DOM_SNAPSHOT_KEY:"domSnapshotResult"};return function(t){return JSON.stringify(a(window[s],l,t))}}()};let n;try{let e=document.head.querySelector("[data-applitools-sandbox]");e||(e=document.createElement("iframe"),e.setAttribute("data-applitools-sandbox",""),e.setAttribute("data-applitools-skip",""),document.head.appendChild(e));const o=window.crypto.getRandomValues(new Uint32Array(1))[0],r=e.contentDocument.createElement("script");r.textContent=`window['ctor-${o}'] = ${t.toString()};`,e.contentDocument.head.appendChild(r),n=e.contentWindow["ctor-"+o]}catch(t){}return n=n||t,n(document).apply(null,arguments)}}();

  return pollResult.apply(this, arguments);
}